

import time
import traceback
import sys
import linecache
import inspect
import os.path
import queue
from io import StringIO
func_code_attr = '__code__'
import subprocess
import threading
import tempfile
import shutil

from mpi4py import MPI

from nose.core import TestProgram
from nose.plugins.capture import Capture
from nose.plugins.skip import Skip, SkipTest
from nose.plugins.doctests import Doctest

from multiprocessing import Process, Queue

from . import project


def is_mpd_running():
        
    name_of_the_vendor, version = MPI.get_vendor()
    if name_of_the_vendor == 'MPICH2':
        must_check_mpd = True
        if 'AMUSE_MPD_CHECK' in os.environ:
            must_check_mpd = os.environ['AMUSE_MPD_CHECK'] == '1'
        if 'PMI_PORT' in os.environ:
            must_check_mpd = False
        if 'HYDRA_CONTROL_FD' in os.environ:
            must_check_mpd = False
        
        if not must_check_mpd:
            return True
        try:
            process = subprocess.Popen(['mpdtrace'], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
            (output_string, error_string) = process.communicate()
            return not (process.returncode == 255)
        except OSError as ex:
            return True
    else:
        return True
        
def ensure_mpd_is_running():
    if not is_mpd_running():
        name_of_the_vendor, version = MPI.get_vendor()
        if name_of_the_vendor == 'MPICH2':
            try:
                process = subprocess.Popen(['nohup','mpd'], close_fds=True)
            except OSError as ex:
                pass



def number_str(number, singular, plural = None):
    if plural == None:
        plural = singular + 's'
    return str(number) + ' ' + (singular if number == 1 else plural)

def find_method_in_class(name_of_the_method, code, class_to_search):
    if name_of_the_method in class_to_search.__dict__:
        member =  class_to_search.__dict__[name_of_the_method]
        if inspect.isfunction(member):
            if getattr(member, func_code_attr) == code:
                return member
        if inspect.ismethoddescriptor(member):
            pass
    return None

def extract_tb(tb, limit = None):
    list = []
    n = 0
    while tb is not None and (limit is None or n < limit):
        f = tb.tb_frame
        lineno = tb.tb_lineno
        co = f.f_code
        filename = co.co_filename
        name = co.co_name
        linecache.checkcache(filename)
        line = ""
        if '__file__' in f.f_globals:
            for global_name, x in f.f_globals.items():
                if global_name.startswith('_'):
                    continue

                if inspect.isfunction(x):
                    if global_name == name and get(x, func_code_attr) == co:
                        args, varargs, varkw, defaults = inspect.getargspec(x)
                        name += inspect.formatargspec(args, varargs, varkw, defaults)
                elif inspect.isclass(x):
                    method = find_method_in_class(name,co,x)
                    if not method is None:
                        args, varargs, varkw, defaults = inspect.getargspec(method)
                        name += inspect.formatargspec(args, varargs, varkw, defaults)
                        name = x.__name__ + '.' + name
                            
        if line:
            line = line.strip()
        else: 
            line = None
        list.append((filename, lineno, name, line))
        tb = tb.tb_next
        n = n+1
    return list


class TestCaseReport(object):
    type = 'unit-report'
    
    def __init__(self, test):
        self.id = test.id()
        self.address = test.address()
        
        self.start_time = 0.0
        self.end_time = 0.0
        self.total_time = 0.0
        self.number_of_runs = 0.0
        
        self.number_of_suite_runs = 0.0
        self.is_paused = False
        
        
        if hasattr(test.test, "_dt_test"):
            self.lineno = test.test._dt_test.lineno
        elif hasattr(test.test, "_testMethodName"):
            method = getattr(test.test, getattr(test.test, "_testMethodName"))
            self.lineno = method.getattr(func_code_attr).co_firstlineno
        else:
            self.lineno = test.test.descriptor.compat_co_firstlineno
            
        self.failed = False
        self.errored = False
        self.skipped = False
        self.found = False
    
    def __str__(self):
        return str(self.id)
    
    def start(self):
        self.start_time = time.time()
        self.end_time = 0.0
        self.found = True
    
    def end(self):
        self.end_time = time.time()
        self.number_of_runs += 1.0
        self.total_time += (self.end_time - self.start_time)
        self.failed = False
        self.errored = False
        self.skipped = False
        
    def mean_time(self):
        if self.number_of_runs == 0:
            return 0.0
        return self.total_time / self.number_of_runs

    def add_syntax_error_file_to_traceback(self, error_type, error_value):
        if not issubclass(error_type, SyntaxError):
            return
            
        try:
            msg, (filename, lineno, offset, badline) = error_value.args
        except Exception:
            pass
        
        self.traceback.insert(0, (filename, lineno, "", ""))
        
    def end_with_error(self, error_tuple):
        self.end_time = time.time()
        error_type, error_value, error_traceback = error_tuple        
        self.error_string = traceback.format_exception_only(error_type, error_value)
        self.traceback = list(reversed(extract_tb(error_traceback)))
        
        self.add_syntax_error_file_to_traceback(error_type, error_value)
        
        self.reset_timing()
        
        self.failed = False
        self.errored = True
        self.skipped = False
        
    def end_with_failure(self, error_tuple):
        self.end_time = time.time()
        error_type, error_value, error_traceback = error_tuple        
        self.error_string = traceback.format_exception_only(error_type, error_value)
        self.traceback = list(reversed(traceback.extract_tb(error_traceback)))
        self.add_syntax_error_file_to_traceback(error_type, error_value)
        
        self.reset_timing()
        
        self.failed = True
        self.errored = False
        self.skipped = False
        
    def end_with_skip(self):
        self.end_time = time.time()
        
        self.skipped = True
        self.failed = False
        self.errored = False
        
    def reset_timing(self):
        self.number_of_runs = 0
        self.total_time = 0.0
            
    def to_dict(self):
        result = self.__dict__.copy()
        result['mean_time'] = self.mean_time()
        return result
        
        
class MakeAReportOfATestRun(object):
    score = 2000
    
    def __init__(self, previous_report = None, reports_queue = None):
        self.errors = 0
        self.failures = 0
        self.tests = 0
        self.start_time = 0
        self.end_time = 0
        self.skipped = 0
        self.report_id = -1
        self.last_test_run = None
        self.crashed = False
        
        if reports_queue is None:
            self._queue_report = self.ignore_report
        else:
            self._queue_report = self.store_report
            self._reports_queue = reports_queue
        
        if previous_report is None:
            self.address_to_report = {}
        else:
            self.address_to_report = previous_report.address_to_report
           
            
        self.name = 'report on a test'
        self.enabled = True

    def __getstate__(self):
        result = {}
        for key, value in self.__dict__.items():
            if not key.startswith('_'):
                result[key] = value
        return result
        
        
    def ignore_report(self, report):
        pass
        
    def store_report(self, report):
        if self.total_number_of_tests() % 1 == 0 or report.failed or report.errored:
            self._reports_queue.put((report.type,report.to_dict(), self.to_information_dict() ))
        
    
    def prepareTest(self, suite):
        pass
        #all_tests = list(suite)
        #number_of_tests = 0
        #self._reports_queue.put(('start-report',{"number_of_tests": number_of_tests}))
        
    def addSuccess(self,test):
        report = self.get_report(test)
        report.end()
        self._queue_report(report)
        
    def addError(self,test,error_tuple):
        report = self.get_report(test)
        error_class, ignore_1, ignore_2 = error_tuple
        if issubclass(error_class, SkipTest):
            self.skipped += 1
            self.tests -= 1
            report.end_with_skip()
        else: 
            report.end_with_error(error_tuple)
            self.errors += 1
        
        self._queue_report(report)
    
    def addFailure(self, test, error_tuple):
        self.failures += 1
        report = self.get_report(test)
        report.end_with_failure(error_tuple)
        
        self._queue_report(report)
           
    def get_report(self, test):
        address = test.address()
        if not address in self.address_to_report:
            self.address_to_report[address] = TestCaseReport(test)
        return self.address_to_report[address] 
        
    def options(self, parser, env):
        pass
        
    def configure(self, parser, env):
        pass
        
        
    def beforeTest(self,test):
        self.tests += 1
        x = self.get_report(test)
        x.start()
        x.number_of_suite_runs += 1
    
    def begin(self):
        self.start_time = time.time()
        self.end_time = 0
        for x in self.address_to_report.values():
            x.errored = False
            x.failed = False
            x.skipped = False
            x.found = False
        
    def finalize(self, x):    
        self.end_time = time.time()
        for key, report in list(self.address_to_report.items()):
            if not report.found:
                del self.address_to_report[key]        

    def startTest(self, test):
        if self.is_test_able_to_run(test):
            report = TestCaseStartReport(test)
            self._queue_report(report)
            return
        else:
            raise SkipTest

    def is_test_able_to_run(self, test):
        report = self.get_report(test)
        if report.is_paused:
            return False
            
        time_taken = report.mean_time()
        if time_taken < 0.1:
            return True
        if time_taken < 0.25:
            return (report.number_of_suite_runs % 5) == 0
        if time_taken < 0.5:
            return (report.number_of_suite_runs % 10) == 0
        if time_taken < 1.0:
            return (report.number_of_suite_runs % 15) == 0
        if time_taken < 3.0:
            return (report.number_of_suite_runs % 30) == 0
            
        return (report.number_of_suite_runs % 40) == 0
    
    def has_errors(self):
        return self.errors > 0
    
    def has_failures(self):
        return self.failures > 0
        
    def has_skipped_tests(self):
        return self.skipped > 0
        
    def total_number_of_tests(self):
        return self.tests + self.skipped  + self.failures + self.errors
        
    def title_string(self):
        title = '';
        if self.has_failures() or self.has_errors():
            title += "FAIL "
        else:
            title += "OK "

        
        if self.has_errors():
            title += 'E' + str(self.errors) + ' '
        
        if self.has_failures():
            title += 'F' + str(self.failures) + ' '
        
        
        title += '+' + str(self.tests) + ' '
        
        if self.has_skipped_tests():
            title += 'S' + str(self.skipped) + ' '
        
        title += 'T' + str(self.total_number_of_tests()) + ' '
        
        if not self.is_running():
            delta_time = self.end_time - self.start_time
            title += ("%10.3f" % delta_time) + 's';
        else:
            cur_time = time.time()
            delta_time =  cur_time - self.start_time
            title += "(running {0:10.3f})".format(delta_time)
            
        title += ' ';
        title += time.strftime("%H:%M:%S", time.gmtime(self.start_time))
        return title;

    def is_running(self):
        return self.end_time <= 1.0
        
    def to_information_dict(self):
        result = {}
        if not self.is_running():
            result['report_id'] = self.report_id
        else:
            result['report_id'] = -1
        result['title'] = self.title_string()
        result['success'] = not self.has_errors() and not self.has_failures()
        return result
    
    def to_dict(self):
        result = self.to_information_dict()
        
        for x in [ 
            'errors', 'failures', 'tests', 
            'start_time', 'end_time', 'skipped',
            ]:
            result[x] = getattr(self, x)
        
        testcases = list(self.address_to_report.values())
        for x in testcases:
            if x.address[0] is None:
                x.address = list(x.address)
                x.address[0] = ''
                
        testcases.sort(key=lambda x: os.path.basename(x.address[0]))
        result['testcases'] = [x.to_dict() for x in testcases]

        return result



class SelectOneTestAndStoreOutput(object):
    name = 'select one test'
    enabled = True
    score = 10
    
    def __init__(self, address):
        self.address = address
        self.capture_stdout = None
        self.stdout = None
        
    def options(self, parser, env):
        pass
        
    def configure(self, parser, env):
        pass
    
    def afterTest(self, test):
        if not self.stdout is None:
            sys.stdout = self.stdout
            self.stdout = None
            
    def startTest(self, test):
        if test.address() == self.address:
            self.stdout = sys.stdout
            self.capture_stdout = StringIO()
            sys.stdout = self.capture_stdout 
            return
        else:
            raise SkipTest
        
    @property
    def buffer(self):
        if self.capture_stdout:
            return self.capture_stdout.getvalue()
        else:
            return "none"

class RunTests(object):
    DIRECTORIES = None
    WORKING_DIRECTORY = None
    
    def __init__(self):
        self.test_is_running = False
        self.report_queue = queue.Queue()
        self.report_info = {}
        self.life_reports = []
        self.get_life_reports_semaphore = threading.Semaphore()
        
    def _perform_the_testrun(self, directories, results_queue, previous_report = None):
        try:
            ensure_mpd_is_running()
            
            null_device = open('/dev/null')
            os.stdin = null_device
            report = MakeAReportOfATestRun(previous_report, results_queue)
            doctest = Doctest()
            doctest.enabled = True
            plugins = [doctest, report, Skip(), Capture()] 
            argv = ['nose', '-v']
            old_working_directory = os.getcwd()
            if not self.WORKING_DIRECTORY is None:
                argv.extend(['-w', self.WORKING_DIRECTORY])
                os.chdir(self.WORKING_DIRECTORY)
                
            argv.extend( directories)
            argv.extend(['--with-doctest', '--doctest-extension=txt'])
            
            result = TestProgram(exit = False, argv=argv, plugins=plugins);
            
            os.chdir(old_working_directory)
            results_queue.put(('test-report', report,) )
        except :
            results_queue.put(('test-error', 'Exception happened: ' + str(sys.exc_info()[0]) + " - " + str(sys.exc_info()[1]), ))
        finally:
            results_queue.put(None)
            MPI.Finalize()


    def _perform_one_test(self, directories, results_queue, address):
        try:
            print("start test run")
            null_device = open('/dev/null')
            os.stdin = null_device
            select = SelectOneTestAndStoreOutput(address)
            plugins = [select, Skip()]  
            argv = ['nose']
            argv.extend( directories)
            result = TestProgram(
                exit = False, 
                argv=argv, 
                plugins=plugins)
            success = result.success
            if success:
                result = 'Success'
            else:
                result = 'Failure'
            
            if select.buffer:
                result += ' - '
                if len(select.buffer) > 1000:
                    result += select.buffer[:min(1000, len(result) - 1)] 
                    result += ' ' + str(result - 1000) + ' more ...'
                else:
                    result += select.buffer
            
            results_queue.put(('test-output',result))
        except:
            results_queue.put(('test-error', 'Exception happened: ' + str(sys.exc_info()[0]) + " - " + str(sys.exc_info()[1]), ))
        finally:
            results_queue.put(None)
            MPI.Finalize()
            print("calling finalize done")

    def run_test_with_address(self, address):
        typestring, result = self.run_in_another_process(self._perform_one_test, address)
        return result
         
    def run_tests(self, previous_report):
        typestring, result = self.run_in_another_process(self._perform_the_testrun, previous_report)
        if typestring == 'test-error':
            print(result)
            raise Exception(result)
        else:
            return result  
            
    def can_start_a_test(self):
        return not self.test_is_running
        
    def run_in_another_process(self, method, argument):
        if self.test_is_running:
            raise Exception("Test is already running")
        
        self.life_reports = []
        self.clear_reports_queue()
        self.test_is_running = True
        self.last_test_started = None
        try:
            result_queue = Queue()
            cwd = os.getcwd()
            if self.DIRECTORIES is None:
                paths = [os.path.abspath(x) for x in project.DIRECTORIES]
            else:
                paths = [os.path.abspath(x) for x in self.DIRECTORIES]
        
            process = Process(
                target=method, 
                args=(
                    paths, 
                    result_queue, 
                    argument
                ))
            
            previous_output = sys.stdout
            previous_error = sys.stderr
            
            temp_stdout = tempfile.NamedTemporaryFile('w')
            temp_stderr = temp_stdout

            print(temp_stdout.name)

            sys.stdout = temp_stdout
            sys.stderr = temp_stderr
            
            process.start()
            
            sys.stdout = previous_output
            sys.stderr = previous_error
            last_message = None
            while True:
                message = result_queue.get(True, 60)
                if message is None:
                    break;
                if message[0] == 'unit-report':
                    print(message[1])
                    self.report_info = message[2]
                    if(message[1]['failed'] or message[1]['errored']):
                        self.report_queue.put(message[1])
                   
                if message[0] == 'start-report':
                    self.last_test_started = message[1]
                    print(self.last_test_started)
                    continue
                    
                last_message = message
            result = last_message
            self.last_test_started = None
        except queue.Empty:
            print("No message recieved from process for 60 seconds")
            report = MakeAReportOfATestRun()
            report.last_test_run = self.last_test_started
            report.start_time = report.end_time = time.time()
            report.crashed = True
            result = 'failed', report
            temp_stdout.flush()
            
            shutil.copyfile(temp_stdout.name, "test_output.txt")
            
            process.terminate()
        finally:
            temp_stdout.close()
            self.test_is_running = False
            self.clear_reports_queue()
            
        process.join(2)
        del result_queue
        return result
     
    def clear_reports_queue(self):
        
        must_clear = not self.report_queue.empty()
        while must_clear:
            try:
                self.report_queue.get_nowait()
                must_clear = not self.report_queue.empty()
            except queue.Empty:
                must_clear = False
                
    def get_reports(self):
        self.get_life_reports_semaphore.acquire()
        try:
            must_clear = not self.report_queue.empty()
            while must_clear:
                try:
                    self.life_reports.append(self.report_queue.get_nowait())
                    must_clear = not self.report_queue.empty()
                except queue.Empty:
                    must_clear = False
            return self.life_reports
        finally:
            self.get_life_reports_semaphore.release()
        
            




class TestCaseStartReport(object):
    type = 'start-report'
    
    def __init__(self, test):
        self.id = test.id()
        self.address = test.address()
        self.start_time = time.time()
            
    def to_dict(self):
        result = self.__dict__.copy()
        return result




from amuse.community import *
from amuse.community.interface.gd import GravitationalDynamicsInterface
from amuse.community.interface.gd import GravitationalDynamics
from numpy import pi
from amuse.units import constants

class MikkolaInterface(CodeInterface,
                       GravitationalDynamicsInterface,
                       StoppingConditionInterface):
    
    use_modules = ['Mikkola', 'StoppingConditions']
    
    def __init__(self, **keyword_arguments):
        CodeInterface.__init__(self, name_of_the_worker="mikkola_worker", **keyword_arguments)
    
    @legacy_function
    def set_time_step():
        """
        Set the model timestep.
        """
        function = LegacyFunctionSpecification()  
        function.addParameter('time_step', dtype='float64', direction=function.IN,
            description = "The current model timestep")
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            Current value of the time step was retrieved
        -1 - ERROR
            The code does not have support for querying the time
        """
        return function
        
    @legacy_function
    def get_lightspeed():
        """
        Get the lightspeed value, the lightspeed scales the units (like G=1)
        and limits the valid velocity terms
        """
        function = LegacyFunctionSpecification()
        function.addParameter('lightspeed', dtype='float64',
                              direction=function.OUT,
            description = "value for the lightspeed")
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            the parameter was retrieved
        -1 - ERROR
            could not retrieve parameter
        """
        return function
        
    @legacy_function
    def set_lightspeed():
        """
        Set the lightspeed value, the lightspeed scales the units (like G=1)
        and limits the valid velocity terms
        """
        function = LegacyFunctionSpecification()
        function.addParameter('lightspeed', dtype='float64',
                              direction=function.IN,
            description = "value for the lightspeed")
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            the parameter was set
        -1 - ERROR
            could not set parameter
        """
        return function
        
    @legacy_function
    def get_tolerance():
        """
        Retrieve the accurancy parameter for the evolve
        """
        function = LegacyFunctionSpecification()
        function.addParameter('tolerance', dtype='float64',
                              direction=function.OUT,
            description = "tolerance for the evolve")
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            the parameter was retrieved
        -1 - ERROR
            could not retrieve parameter
        """
        return function
        
    @legacy_function
    def set_tolerance():
        """
        Set the accurancy parameter for the evolve
        """
        function = LegacyFunctionSpecification()
        function.addParameter('tolerance', dtype='float64',
                              direction=function.IN,
            description = "tolerance for the evolve")
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            the parameter was set
        -1 - ERROR
            could not set parameter
        """
        return function
    
    @legacy_function
    def get_maximum_number_of_particles():
        """
        Retrieve the maximum number of particles that can be evolved with this code
        """
        function = LegacyFunctionSpecification()
        function.addParameter(
            'maximum_number_of_particles', 
            dtype='int32',
            direction=function.OUT
        )
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            the parameter was retrieved
        -1 - ERROR
            could not retrieve parameter
        """
        return function
        
    @legacy_function
    def set_maximum_number_of_particles():
        """
        Change the maximum number of particles that can be evolved with this code
        """
        function = LegacyFunctionSpecification()
        function.addParameter(
            'maximum_number_of_particles', 
            dtype='int32',
            direction=function.IN
        )
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            the parameter was set
        -1 - ERROR
            could not set parameter
        """
        return function
    
    @legacy_function
    def get_evolve_to_exact_time():
        """
        Evolve to model to the exact time given in the evolve_model call (can be slower)
        """
        function = LegacyFunctionSpecification()
        function.addParameter(
            'evolve_to_exact_time', 
            dtype='bool',
            direction=function.OUT
        )
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            the parameter was retrieved
        -1 - ERROR
            could not retrieve parameter
        """
        return function
        
    @legacy_function
    def set_evolve_to_exact_time():
        """
        Evolve to model to the exact time given in the evolve_model call (can be slower)
        """
        function = LegacyFunctionSpecification()
        function.addParameter(
            'evolve_to_exact_time', 
            dtype='bool',
            direction=function.IN
        )
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            the parameter was set
        -1 - ERROR
            could not set parameter
        """
        return function
    
    @legacy_function
    def get_radiated_gravitational_energy():
        """
        Retrieve the current radiated gravitational energy of the model
        """
        function = LegacyFunctionSpecification()
        function.addParameter('radiated_gravitational_energy', dtype='float64', direction=function.OUT,
            description = "The energy radiated by gravitational waves")
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            Current value of the energy was set
        -1 - ERROR
            Energy could not be provided
        """
        return function
    
    
    @legacy_function
    def get_total_energy():
        """
        Retrieve the current total energy of the model
        """
        function = LegacyFunctionSpecification()
        function.addParameter('energy', dtype='float64', direction=function.OUT,
            description = "The energy radiated by gravitational waves")
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            Current value of the energy was set
        -1 - ERROR
            Energy could not be provided
        """
        return function
        
    @legacy_function
    def get_number_of_particles_added():
        """
        Return the number of particles added during the last evolve.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function
    
    @legacy_function
    def get_id_of_added_particle():
        """
        Return the id of the new particle in the code
        """
        function = LegacyFunctionSpecification()
        function.addParameter('index_of_add', dtype='int32',
                              direction=function.IN, 
                 description = 'index in the added particles list (0-n)')
        function.addParameter('index_of_particle', dtype='int32',
                              direction=function.OUT)
        function.can_handle_array = True
        function.result_type = 'int32'
        return function
    
    @legacy_function
    def get_children_of_particle():
        """
        Return the number of particles added or deleted during the last evolve.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('index_of_the_particle', dtype='int32',
                              direction=function.IN, 
                 description = 'index of the parent particle',
                 unit = INDEX)
        function.addParameter('child1', dtype='int32', direction=function.OUT,
                description = 'index of the first child particle, -1 if none',
                unit = LINK('particles') )
        function.addParameter('child2', dtype='int32', direction=function.OUT,
                unit = LINK('particles'))
        function.can_handle_array = True
        function.result_type = 'int32'
        return function
        
class Mikkola(GravitationalDynamics):

    def __init__(self, convert_nbody=None, **options):
      
        self.stopping_conditions = StoppingConditions(self)
        GravitationalDynamics.__init__(
            self, 
            MikkolaInterface(**options),
            convert_nbody,
            **options
        )
    
    def initialize_code(self):
        result = self.overridden().initialize_code()
        
        if not self.unit_converter is None:
            value=self.unit_converter.to_nbody(constants.c)
            self.parameters._original.lightspeed = value

        return result
        
    def define_parameters(self, handler):
        #~ GravitationalDynamics.define_parameters(self, handler)
        
        handler.add_method_parameter(
            "get_time_step",
            "set_time_step",
            "timestep",
            "initial timestep for iteration", 
            default_value = 1.0 | nbody_system.time
        )
        handler.add_method_parameter(
            "get_lightspeed", 
            "set_lightspeed",
            "lightspeed", 
            "lightspeed used in the code", 
            default_value = 1.0 | nbody_system.length / nbody_system.time
        )
        handler.add_method_parameter(
            "get_tolerance", 
            "set_tolerance",
            "tolerance", 
            "tolerance used in the code", 
            default_value = 1e-13
        )
        handler.add_method_parameter(
            "get_maximum_number_of_particles", 
            "set_maximum_number_of_particles",
            "maximum_number_of_particles", 
            "the code will evolve this number of particles, please be sure to account for mergers", 
            default_value = 100
        )
        handler.add_boolean_parameter(
            "get_evolve_to_exact_time", 
            "set_evolve_to_exact_time",
            "evolve_to_exact_time", 
            "the code will evolve the model to the exact time given in evolve_model", 
            True
        )
        
    def define_methods(self, handler):
        GravitationalDynamics.define_methods(self, handler)
        
        handler.add_method(
            "set_lightspeed",
            ( nbody_system.length / nbody_system.time, ),
            (handler.ERROR_CODE,)
        )
        
        handler.add_method(
            "get_lightspeed",
            ( ),
            (nbody_system.length / nbody_system.time, handler.ERROR_CODE,)
        )
        
        handler.add_method(
            "set_time_step",
            (nbody_system.time, ),
            (handler.ERROR_CODE,)
        )

        handler.add_method(
            "set_tolerance",
            (handler.NO_UNIT, ),
            (handler.ERROR_CODE,)
        )
        
        handler.add_method(
            "get_tolerance",
            ( ),
            (handler.NO_UNIT, handler.ERROR_CODE,)
        )
        
        handler.add_method(
            "get_radiated_gravitational_energy",
            (),
            (nbody_system.mass * nbody_system.length ** 2  * nbody_system.time ** -2, handler.ERROR_CODE,)
        )


        handler.add_method(
            "get_total_energy",
            (),
            (nbody_system.mass * nbody_system.length ** 2  * nbody_system.time ** -2, handler.ERROR_CODE,)
        )

    def define_properties(self, handler):
        
        GravitationalDynamics.define_properties(self, handler)
        handler.add_property("get_radiated_gravitational_energy")
    
    def update_particle_set(self):
        """
        update the particle set after changes in the code
        
        this implementation needs to move to the
        amuse.datamodel.incode_storage module, as
        it uses a lot of internal methods and info!
        
        """  
        
        number_of_added_particles = self.get_number_of_particles_added()
        if number_of_added_particles == 0:
            return
        
        indices_in_update_list = list(range(number_of_added_particles))
        indices_to_add = self.get_id_of_added_particle(indices_in_update_list)
        
        incode_storage = self.particles._private.attribute_storage
        
        if len(indices_to_add) > 0:
            incode_storage._add_indices(indices_to_add)


    def define_particle_sets(self, handler):
        GravitationalDynamics.define_particle_sets(self, handler)
        self.stopping_conditions.define_particle_set(handler)
        





import subprocess
import socket, os

import time
import urllib.parse
import threading
import json
import os
import sys
import re

import webbrowser

from xml.dom import minidom

from optparse import OptionParser

from . import webserver
from . import background_test
from . import project
import pickle
import textwrap

from queue import Queue

background_test.RunTests.instance = background_test.RunTests()

class late(object):
    def __init__(self, initializer):
        
        self.initializer = initializer
        self.__doc__ = self.initializer.__doc__
        
    def __get__(self, instance, owner):
        if instance is None:
            return self
        value = self.initializer(instance)
        setattr(instance,self.initializer.__name__, value)
        return value
        
class SendAnEmail(object):
    
    def __init__(self, **keyword_arguments):
        if len(keyword_arguments) > 0:
            for key, value in keyword_arguments.items():
                setattr(self, key, value)
                self.start()

    def start(self):
        
        call = ['mail','-s',self.subject, '-r', self.sender_email_address]
        call.extend(self.recipients)
        
        print(call)
        
        process = subprocess.Popen(
            call,
            stdout=subprocess.PIPE,
            stdin =subprocess.PIPE,
            stderr=subprocess.PIPE
        )
            
        stdout, stderr = process.communicate(self.mail_contents)
        
        if process.returncode != 0:
            raise Exception("Could not send e-mail, error output was {0}".format(stderr))
        
    @late
    def sender_email_address(self):
        if "EMAIL" in os.environ:
            return os.environ["EMAIL"]
        else:
            return 'noreply@{0}'.format(socket.getfqdn())
    
    @late
    def mail_contents(self):
        return 'Mail send by the SendAnEmail class\nContents not provided\n'
    
    @late
    def subject(self):
        return 'Automatic mail subject'
        


def get_first_element_with_tag(parent, name):
    for node in parent.childNodes:
        if node.nodeType == minidom.Node.ELEMENT_NODE and \
            (name == "*" or node.tagName == name):
            return node
    return None
    
header = """\
Dear {name},

"""

errored_start = """\
This is to inform you that your commit had errors.
"""
success_start = """\
This is to inform you that your commit had no errors, well done!
"""
crash_start = """\
This is to inform you that during testing of your commit the code crashed.

The last test run was: {last_test_run}
"""

commit_info = """\
AUTHOR   : {author}
REVISION : {revision}
DATE     : {date}
MESSAGE  : {msg}
"""

error_info = """\
NUMBER OF ERRORS        : {number_of_errors:>5d}
"""

skip_info = """\
NUMBER OF SKIPPED TESTS : {number_of_skips:>5d}
"""

tests_info = """\
NUMBER OF TESTS         : {number_of_tests:>5d}
TIME TAKEN(seconds)     : {number_of_seconds:>6.1f}
"""
footer = """\
Regards,

The AMUSE automated testing system
"""

more_tests = """\
The number of tests has increased with : {delta_tests}. Good job!
"""
less_tests = """\
Watch out, the number of tests has decreased! There are now {0} less tests. 
"""
same_tests = """\
The number of tests has not increased, is your code tested?
"""

crashed_email_subject = """Test {last_test_run} crashed in revision {revision}"""
errored_email_subject = """Found {number_of_errors} error(s) in revision {revision}"""
success_email_subject = """For revision revision {revision}, all {number_of_tests} tests were successful!"""


def run_command(arguments):
    print("running :" + ' '.join(arguments))
    process = subprocess.Popen(arguments, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    stdoutstr, stderrstr = process.communicate()
    print(stderrstr)
    return stdoutstr

class RequestACodeReview(object):
    
    def __init__(self, **keyword_arguments):
        if len(keyword_arguments) > 0:
            for key, value in keyword_arguments.items():
                setattr(self, key, value)
                self.start()

    def start(self):
        if not self.log_string:
            pass
        
        if subprocess.call(['which', 'post-review']) != 0:
            return
            
        revision = self.revision
        
        repository_url =  '--repository-url=' + self.svn_repository_url
        password = '--password=' + self.reviewboard_password
        username = '--username=' + self.reviewboard_username
        description = "--description=(In [%s]) %s" % (revision, self.log_string)
        submitas = '--submit-as=' + self.author
        revision = '--revision-range=%s:%s' % (int(revision)-1, revision)
        server = '--server=' + self.reviewboard_url
        
        arguments = [
            'post-review',
            repository_url,
            password,
            username,
            submitas,
            revision,
            server,
            self.review_identifier_argument,
            self.publish_argument,
            self.testing_done_argument
        ]
        
        if len(self.review_identifier_argument) == 0:
            arguments += [self.summary_argument, description]
        
        print(run_command(arguments))
        
    @late
    def svn_repository_url(self):
        return 'http://www.amusecode.org/svn'
    
    @late
    def reviewboard_url(self):
        return 'http://www.amusecode.org/review/'
        
    @late
    def reviewboard_username(self):
        return 'svn'
        
    @late
    def reviewboard_password(self):
        return os.environ['RVPW']

    @late
    def author(self):
        return 'unknown'
    
    @late
    def log_string(self):
        return ''
    
    @late
    def review_identifier_argument(self):
        m = re.search(r'update(?: )?review:([0-9]+)', self.log_string, re.M | re.I)
        if m:
            return '--review-request-id=' + m.group(1)
        else:
            return ''
            
    @late
    def publish_argument(self):
        if re.search(r'draft(?: )?review', self.log_string, re.M | re.I):
            return ''
        else:
            return '-p'
    
    @late
    def summary_argument(self):
        return '--summary=' + self.log_string[:250].splitlines().pop(0).split('. ').pop(0)
    
    @late
    def testing_done_argument(self):
        if self.test_report:
            return '--testing-done=' + self.test_report
        else:
            return ''



            
class RunAllTestsOnASvnCommit(object):
    DEFAULT = None
    
    def __init__(self):
        self.queue =  Queue()
        self.must_run = False

    @classmethod
    def default(cls):
        if cls.DEFAULT is None:
            cls.DEFAULT = cls()
        return cls.DEFAULT
    
    @late
    def working_directory(self):
        path = os.path.dirname(os.path.dirname(__file__))
        path = os.path.abspath(os.path.join(path, 'working-copy'))
        return path
    
    @late
    def directories(self):
        return [os.path.join(self.working_directory, x) for x in project.DIRECTORIES]
    
    @late
    def mapping_from_author_to_email(self):
        path = os.path.dirname(__file__)
        path = os.path.join(path, "authors.map")
        with open(path, 'r') as f:
            return pickle.load(f)
    
    def cleanup_compiled_python_files(self):
        for path in self.find_pyc_files(self.working_directory):
            os.remove(path)

    def find_pyc_files(self, rootname):
        for dirname, subdirectories, files in os.walk(rootname):
            for filename in files:
                if filename.endswith('.pyc'):
                    yield os.path.join(dirname, filename)


    
    def update_from_svn(self, revision):
        subprocess.call(['svn', '--force', 'update', '-r', revision], cwd = self.working_directory)
        
    def build_code(self):
        subprocess.call(['make','clean'], cwd = self.working_directory)
        subprocess.call(['make'], cwd = self.working_directory)
        
    def get_author_date_and_msg_for(self, revision):
        process = subprocess.Popen(['svn','log', '-r', revision, '--xml'], cwd = self.working_directory, stdout = subprocess.PIPE)
        result, ignore = process.communicate()
        print(result)
        if not process.returncode == 0:
            raise Exception("could not retrieve log for revision {0}" + revision)
            
        doc = minidom.parseString(result)
        results = []
        entry = list(doc.getElementsByTagName('logentry'))[0]
        author  =  get_first_element_with_tag(entry, "author").firstChild.data
        date_string = get_first_element_with_tag(entry, "date").firstChild.data
        msg_string = get_first_element_with_tag(entry, "msg").firstChild.data
            
        return author, date_string, msg_string
        
    def run_all_tests(self):
        background_test.RunTests.DIRECTORIES = self.directories
        background_test.RunTests.WORKING_DIRECTORY = self.working_directory
        return background_test.RunTests.instance.run_tests(None)
        
    def send_report_as_email_to(self, report, mail_content_string, recipient):
        uc = SendAnEmail()
        uc.mail_contents = mail_content_string
        
        
        if report["crashed"]:
            uc.subject = crashed_email_subject.format(**report)
        elif report["number_of_errors"] > 0:
            uc.subject = errored_email_subject.format(**report)
        else:
            uc.subject = success_email_subject.format(**report)
        
        if not recipient is None:
            uc.recipients = [recipient]
            uc.start()
        
        uc.recipients = [self.admin_email_address]
        uc.start()
        
    def new_mail_content_string(self, report):
        contents = []
        contents.append(header.format(**report))
        
        if report["crashed"]:
            contents.append(crash_start.format(**report))
        elif report["number_of_errors"] > 0:
            contents.append(errored_start.format(**report))
        else:
            contents.append(success_start.format(**report))
            
        
        if report['delta_tests'] > 0:
            contents.append(more_tests.format(**report))
        elif report['delta_tests'] < 0:
            contents.append(less_tests.format(-1 * report['delta_tests']))
        else:
            contents.append(same_tests.format(**report))
            
        contents.append(commit_info.format(**report))
        
        if report["number_of_errors"] > 0:
            contents.append(error_info.format(**report))
            
        if report["number_of_skips"] > 0:
            contents.append(skip_info.format(**report))
            
            
        contents.append(tests_info.format(**report))
            
        
        if report['errors']:
            for location_line, error_string in report['errors']:
                contents.append(location_line)
                contents.append(textwrap.fill('\n'.join(error_string), 80, initial_indent = "  "))
            contents.append('')
                
        contents.append(footer.format(**report))
        
        return '\n'.join(contents)
        
    def check_svn_commit(self,revision):
        
        self.cleanup_compiled_python_files()
        
        
        author, date, msg = self.get_author_date_and_msg_for(revision)
        if author in self.mapping_from_author_to_email:
            name, email = self.mapping_from_author_to_email[author]
        else:
            name, email = 'Admin', None
        
        self.update_from_svn(revision)
        self.build_code()
        
        test_report = self.run_all_tests()
        
        report = {}
        report['author'] = author
        report['date'] = date
        report['msg'] = msg
        report['name'] = name
        report['number_of_errors'] = test_report.errors + test_report.failures
        report['number_of_skips'] = test_report.skipped
        report['number_of_tests'] = test_report.tests
        report['number_of_seconds'] = test_report.end_time - test_report.start_time
        report['revision'] = revision
        report['crashed'] = test_report.crashed
        report['last_test_run'] = test_report.last_test_run
        
        previous_revision = int(revision) - 1
        while not previous_revision in self.mapping_from_revision_to_report and previous_revision > 0:
            previous_revision -= 1
            if previous_revision in self.mapping_from_revision_to_report:
                previous_report = self.mapping_from_revision_to_report[previous_revision]
                if previous_report['crashed']:
                    previous_revision -= 1
           
        if previous_revision in self.mapping_from_revision_to_report:
            previous_report = self.mapping_from_revision_to_report[previous_revision]
            previous_number_of_tests = previous_report['number_of_tests']
            delta_in_number_of_tests = test_report.tests - previous_number_of_tests
        else:
            delta_in_number_of_tests = 0
            
        report['delta_tests'] = delta_in_number_of_tests
        
        
        testcases = list(test_report.address_to_report.values())
        testcases.sort(key=lambda x: os.path.basename("" if x.address[0] is None else x.address[0]))
       
      
        errors = []
        for x in testcases:
            if x.errored or x.failed:
                if not x.address[0] is None:
                    filename = os.path.basename(x.address[0])
                else:
                    filename = 'unknown-file'
                
                location_line = "{0}:{1} {2}".format(filename, x.lineno, x.address[2])
                error_string = x.error_string
                
                errors.append((location_line,error_string,))
                
        report['errors'] = errors
        
        self.mapping_from_revision_to_report[int(revision)] = report
        self.dump_revision_reports()
        
        content = self.new_mail_content_string(report)
        self.send_report_as_email_to(report, content,  email)
        
        requestACodeReview = RequestACodeReview()
        requestACodeReview.log_string = msg
        requestACodeReview.author = author
        requestACodeReview.revision = revision
        requestACodeReview.test_report = content
        requestACodeReview.start()
        
        
    @late
    def admin_email_address(self):
        if "EMAIL" in os.environ:
            return os.environ["EMAIL"]
        else:
            return None
            
    def start(self):
        self.thread = threading.Thread(target = self.runloop)
        self.thread.daemon = True
        self.must_run = True
        self.thread.start()
    
          
    def runloop(self): 
        while self.must_run:
            revision = self.queue.get()
            if revision is None:
                self.must_run = False
            else:
                self.check_svn_commit(revision)
        
        
    def queue_check(self, revision):
        self.queue.put(revision)
        
    def stop(self):
        self.queue.put(None)
    
    def number_of_tested_revisions(self):
        return len(self.mapping_from_revision_to_report)
        
    def dump_revision_reports(self):
        path = os.path.dirname(__file__)
        path = os.path.join(path, "reports.map")
        with open(path, 'w') as f:
            pickle.dump(self.mapping_from_revision_to_report, f)
    
    @late
    def mapping_from_revision_to_report(self):
        path = os.path.dirname(__file__)
        path = os.path.join(path, "reports.map")
        if os.path.exists(path):
            try:
                with open(path, 'r') as f:
                    return pickle.load(f)
            except IOError:
                return {}
        else:
            return {}
        
        
class HandleRequest(webserver.HandleRequest):
   
    def do_check_svn_commit(self):
        parameters = urllib.parse.parse_qs(self.parsed_path.query)
        revision = parameters['rev'][0]
        
        
        self.server.tracker.queue_check(revision)
        
        string = json.dumps("test for revision {0} queued".format(revision))
        content_type = 'text/javascript'
        return string, content_type
        
    
    def do_ping(self):
        string = json.dumps(True)
        content_type = 'text/javascript'
        return string, content_type
        
    def index_file(self):
        if True:
            return ("nothing here", "text/html")
            
        base = os.path.split(__file__)[0]
        filename = os.path.join(base, "tracker.html")
        with open(filename, "r") as file:
            contents = file.read()
        return contents, 'text/html'
            

class ContinuosTestWebServer(webserver.WebServer):
    
    def __init__(self, port):
        webserver.WebServer.__init__(self,  port, HandleRequest)
        self.tracker = RunAllTestsOnASvnCommit()
        
        
    def stop(self):
        self.tracker.stop()
        self.shutdown()
        
            
if __name__ == '__main__':
    parser = OptionParser() 
    
    parser.add_option("-p", "--port", 
      dest="serverport",
      help="start serving on PORT", 
      metavar="PORT", 
      default=9075,
      type="int")
    
    parser.add_option("-a", "--admin", 
      dest="admin_email",
      help="e-mail address of the admin", 
      default=None,
      type="string")
      
    (options, args) = parser.parse_args()
    
    print("starting server on port: ", options.serverport)
      
    server = ContinuosTestWebServer(options.serverport)
    
    if options.admin_email:
        server.tracker.admin_email_address = options.admin_email
    else:
        parser.error("Must set the admin e-mail address using -a")
        
    server.tracker.start()
    server.start()

    

        

        


    
    
